#include <sstream>
#include "CompoundType.h"
#include"ByteCode.h"

using namespace ls;

shared_ptr<CompoundType> ls::CompoundType::binaryGo(Opers op, CompoundType * ptr)
{
	switch (op) {
	case kEqual:
		return createVar(this == ptr);
		break;
	case kNotEqual:
		return createVar(this != ptr);
		break;
	default:
		return nullptr;
		break;
	}
}

shared_ptr<CompoundType> ls::BooleanType::unaryGo(Opers op)
{
	if (op == kNot) {
		return createVar(!value);
	}
	return nullptr;
}

shared_ptr<CompoundType> ls::BooleanType::binaryGo(Opers op, CompoundType * rhs)
{
	auto r = dynamic_cast<BooleanType *>(rhs);
	if (r == nullptr) return nullptr;
	auto ret = LogicalBinaryGo<bool, bool>(op, value, r->value);
	if (ret != nullptr) return ret;
	switch (op) {
	case kEqual:
		return createVar(value == r->value);
		break;
	case kNotEqual:
		return createVar(value != r->value);
		break;
	default:
		return nullptr;
		break;
	}
}

shared_ptr<CompoundType> ls::BooleanType::copy()
{
	return createVar(value);
}

shared_ptr<CompoundType> ls::IntegerType::unaryGo(Opers op)
{
	return ArithmeticUnaryGo(op, value);
}

shared_ptr<CompoundType> ls::IntegerType::binaryGo(Opers op, CompoundType * rhs)
{
	switch (rhs->Type()) {
	case kInteger:
		return ArithmeticBinaryGo(op, value, dynamic_cast<IntegerType *>(rhs)->value);
		break;
	case kFloat:
		return ArithmeticBinaryGo(op, value, dynamic_cast<FloatType *>(rhs)->value);
		break;
	case kString:
		if (op== kAdd){
			return createVar(to_string() + dynamic_cast<StringType *>(rhs)->value);
		}
		break;
	}
	return nullptr;
}

shared_ptr<CompoundType> ls::IntegerType::copy()
{
	return createVar(value);
}

shared_ptr<CompoundType> ls::FloatType::unaryGo(Opers op)
{
	return ArithmeticUnaryGo(op, value);
}

shared_ptr<CompoundType> ls::FloatType::binaryGo(Opers op, CompoundType * rhs)
{
	switch (rhs->Type()) {
	case kInteger:
		return ArithmeticBinaryGo(op, value, dynamic_cast<IntegerType *>(rhs)->value);
		break;
	case kFloat:
		return ArithmeticBinaryGo(op, value, dynamic_cast<FloatType *>(rhs)->value);
		break;
	default:
		return nullptr;
		break;
	}
}

shared_ptr<CompoundType> ls::FloatType::copy()
{
	return createVar(value);
}

shared_ptr<CompoundType> ls::StringType::binaryGo(Opers op, CompoundType * rhs)
{
	if (op != kAdd)
		return nullptr;
	return createVar(value + rhs->to_string());
}

shared_ptr<CompoundType> ls::StringType::copy()
{
	return createVar(value);
}

std::string ls::TableType::to_string()
{
	std::stringstream ss;
	ss << "{";
	for (auto& it : map){
		ss << it.first->to_string() << "="
			<< it.second->to_string() << ",";
	}
	ss << "}";
	return ss.str();
}

void ls::FunctionCodeDeleter::operator()(FunctionCode * p)
{
	delete p;
}
