#include<string>
#include<cstdint>
#include<memory>
#include<unordered_map>
#include<exception>
#include<functional>
#include"Operators.h"

#ifndef LS_CompoundType_H_
#define LS_CompoundType_H_
namespace ls {
	using std::shared_ptr;
	using std::make_shared;

	// ٲ ÿ.
	enum TypeKind {
		kNone = 0,
		kBoolean,
		kInteger,
		kFloat,
		kString,
		kTable,
		kTableIter,
		kFunction,
		kFunctionProxy,
		kUserDefined,
		kDEF_END
	};
	class BoxedVar;
	struct FunctionCode;
	//std::unordered_map<std::string, void(IntType::*)()> st;
	class TypeError : public std::logic_error
	{
	public:
		using std::logic_error::logic_error;
	};
	class InvalidOperation : public TypeError
	{
	public:
		InvalidOperation() :TypeError("InvalidOperation") {}
	};
	class DividedByZero : public TypeError
	{
	public:
		DividedByZero() :TypeError("attempt to divide by zero") {}
	};

	class CompoundType
	{
	public:
		CompoundType() = default;

		//virtual void searchMethod(const std::string & str) {}
		virtual BoxedVar * index(CompoundType *) { return nullptr; }
		
		//ٲٴ  ؾ.
		virtual bool next() { return false; }
		virtual shared_ptr<CompoundType> unaryGo(Opers op) { return nullptr; }
		virtual shared_ptr<CompoundType> binaryGo(Opers op, CompoundType *);

		virtual TypeKind Type() = 0;

		virtual shared_ptr<CompoundType> copy() { return nullptr; }

		//virtual std::shared_ptr<CompoundType> cast(TypeKind tk) { throw InvalidOperation(); }
		virtual std::string to_string() { return "<UnknownObject>"; }
		virtual ~CompoundType() {}

		virtual size_t hash() { return reinterpret_cast<size_t>(this); }
	};
}
namespace std {
	template<> struct hash<ls::CompoundType *>
	{
		typedef ls::CompoundType * argument_type;
		typedef std::size_t result_type;
		result_type operator()(argument_type s) const noexcept{
			return s->hash();
		}
	};
}
namespace ls
{
	class NoneType final : public CompoundType {
	public:
		virtual TypeKind Type() { return TypeKind::kNone; }
		virtual std::string to_string() { return "None"; }
		virtual size_t hash() override { return 0x12345678; };
	};
	class BooleanType final : public CompoundType {
	public:
		explicit BooleanType(bool b) :value(b) {}
		virtual TypeKind Type() override { return TypeKind::kBoolean; }
		virtual std::string to_string() { return value ? "true" : "false"; }
		virtual shared_ptr<CompoundType> unaryGo(Opers op) override;
		virtual shared_ptr<CompoundType> binaryGo(Opers op, CompoundType * rhs) override;
		virtual size_t hash() override { return std::hash<bool>{}(value); };
		virtual shared_ptr<CompoundType> copy();

		bool value;
	};

	class IntegerType final : public CompoundType {
	public:
		explicit IntegerType(int64_t i) :value(i) {}
		virtual TypeKind Type() { return TypeKind::kInteger; }
		virtual std::string to_string() { return std::to_string(value); }
		virtual shared_ptr<CompoundType> unaryGo(Opers op) override;
		virtual shared_ptr<CompoundType> binaryGo(Opers op, CompoundType * rhs) override;
		virtual size_t hash() override { return value; };
		virtual shared_ptr<CompoundType> copy();

		int64_t value;
	};

	class FloatType final : public CompoundType {
	public:
		explicit FloatType(double i) :value(i) {}
		virtual TypeKind Type() { return TypeKind::kFloat; }
		virtual std::string to_string() { return std::to_string(value); }

		virtual shared_ptr<CompoundType> unaryGo(Opers op) override;
		virtual shared_ptr<CompoundType> binaryGo(Opers op, CompoundType * rhs) override;
		virtual size_t hash() override { return std::hash<double>{}(value); };
		virtual shared_ptr<CompoundType> copy();

		double value;
	};

	class StringType final : public CompoundType {
	public:
		explicit StringType(const std::string & str) :value(str) {}
		virtual TypeKind Type() { return TypeKind::kString; }
		virtual std::string to_string() { return "\"" + value + "\""; }

		virtual shared_ptr<CompoundType> unaryGo(Opers op) override { return nullptr; }
		virtual shared_ptr<CompoundType> binaryGo(Opers op, CompoundType * rhs) override;
		virtual size_t hash() override { return std::hash<std::string>{}(value); };
		virtual shared_ptr<CompoundType> copy();

		std::string value;
	};

	typedef std::unordered_map< CompoundType *, shared_ptr<CompoundType> > dict_type;
	typedef dict_type::iterator dict_iterator_type;
	
	/*class IterativeType : public CompoundType {
	};
	ʿұ?
	*/
	class TableIterType final : public CompoundType {
	public:
		explicit TableIterType(dict_iterator_type it, dict_iterator_type end_iterator)
			:cur(it), end(end_iterator) {}
		virtual TypeKind Type() { return TypeKind::kTableIter; }
		virtual std::string to_string() override {
			if (cur != end)
				return cur->first->to_string();
			else return "<End Iteraor>";
		}
		inline virtual bool next() override {
			if (cur != end)
				cur++;
			return true;
		}
		inline bool isEnd() {
			return cur != end;
		}
		inline shared_ptr<CompoundType> & value() {
			return cur->second;
		}
	private:
		dict_iterator_type cur;
		dict_iterator_type end;
	};
	class TableType final : public CompoundType {
	public:
		explicit TableType() {}
		virtual TypeKind Type() { return TypeKind::kTable; }
		virtual std::string to_string() override;
		virtual shared_ptr<TableIterType> getiter() {
			return make_shared<TableIterType>(map.begin(), map.end());
		}
	private:
		dict_type map;
	};
	class FunctionProxyType : public CompoundType {
	public:
		explicit FunctionProxyType(std::function<std::shared_ptr<CompoundType>(std::vector<BoxedVar>&)> v) :value(v) {}
		std::shared_ptr<CompoundType> call(std::vector<BoxedVar> & stack) {
			return value(stack);
		}
		virtual TypeKind Type() override { return TypeKind::kFunctionProxy; }
		virtual std::string to_string() override { return debugName; }
		std::string debugName = "<FunctionProxy>";
		std::function<std::shared_ptr<CompoundType>(std::vector<BoxedVar>&)> value;
	};

	struct FunctionCodeDeleter{
		void operator()(FunctionCode * );
	};

	class FunctionType : public CompoundType {
	public:
		FunctionType(const std::string & n, std::unique_ptr<FunctionCode, FunctionCodeDeleter> c)
			:name(n), code(move(c)) {}

		virtual TypeKind Type() override { return TypeKind::kFunction; }
		virtual std::string to_string() override { return "<Function:"+name+">"; }
	public:
		std::string name = "<Function>";
		std::unique_ptr<FunctionCode, FunctionCodeDeleter> code;
	};
	inline std::shared_ptr<CompoundType> createVar(void) {
		static std::shared_ptr < CompoundType > ret(new NoneType);
		return ret;
	}
	inline std::shared_ptr<CompoundType> createVar(int64_t t) {
		return std::make_shared<IntegerType>(t);
	}
	inline std::shared_ptr<CompoundType> createVar(int t) {
		return std::make_shared<IntegerType>(t);
	}
	inline std::shared_ptr<CompoundType> createVar(bool t) {
		return std::make_shared<BooleanType>(t);
	}
	inline std::shared_ptr<CompoundType> createVar(double t) {
		return std::make_shared<FloatType>(t);
	}
	inline std::shared_ptr<CompoundType> createVar(const std::string & t) {
		return std::make_shared<StringType>(t);
	}
	inline std::shared_ptr<CompoundType> createVar(std::function<std::shared_ptr<CompoundType>(std::vector<BoxedVar>&)> v) {
		return std::make_shared<FunctionProxyType>(v);
	}
	inline std::shared_ptr<CompoundType> createVar(const std::string & name, std::unique_ptr<FunctionCode, FunctionCodeDeleter> t) {
		return std::make_shared<FunctionType>(name, move(t));
	}
	template<typename LHStype,typename RHStype>
	shared_ptr<CompoundType> LogicalBinaryGo(Opers op,const LHStype & lhs,const RHStype & rhs) {
		switch (op)
		{
		case kAnd:
			return createVar(lhs && rhs);
			break;
		case kOr:
			return createVar(lhs || rhs);
			break;
		default:
			return nullptr;
			break;
		}
	}
	template<typename LHStype>
	shared_ptr<CompoundType> ArithmeticUnaryGo(Opers op, const LHStype & lhs) {
		switch (op)
		{
		case kPlus:
			return createVar(+lhs);
			break;
		case kMinus:
			return createVar(-lhs);
			break;
		default:
			return nullptr;
			break;
		}
	}
	template<typename LHStype, typename RHStype>
	shared_ptr<CompoundType> ArithmeticBinaryGo(Opers op, const LHStype & lhs, const RHStype & rhs) {
		switch (op)
		{
		case kAdd:
			return createVar(lhs + rhs);
			break;
		case kSub:
			return createVar(lhs - rhs);
			break;
		case kMul:
			return createVar(lhs * rhs);
			break;
		case kDiv:
			if (rhs == 0){
				throw DividedByZero();
			}
			return createVar(lhs / rhs);
			break;
		case kGreater:
			return createVar(lhs > rhs);
			break;
		case kEqualGreater:
			return createVar(lhs >= rhs);
			break;
		case kEqual:
			return createVar(lhs == rhs);
			break;
		case kNotEqual:
			return createVar(lhs != rhs);
			break;
		case kLess:
			return createVar(lhs < rhs);
			break;
		case kLessEqual:
			return createVar(lhs <= rhs);
			break;
		default:
			return nullptr;
			break;
		}
	}
}
#endif