#include<algorithm>
#include "Lexer.h"

using namespace ls;

static bool isCutCharactor(int ch) {
	static char s[] = " \n\t+-/*.,!|<>[]{}()&\'\"^%:;~`$#\\?#@=\r\0";
	auto pos = std::find(std::begin(s), std::end(s), ch);
	return pos != std::end(s);
}

static bool isOperatorWord(int ch) {
	static char s[] = "+-/*!|<>&^%~$?=(),";
	auto pos = std::find(std::begin(s), std::end(s), ch);
	return pos != std::end(s);
}

static bool isIgnoreWord(int ch) {
	static char s[] = " \n\r\t";
	auto pos = std::find(std::begin(s), std::end(s), ch);
	return pos != std::end(s);
}

Token IstreamTokenizer::get()
{
	Token t;
	if (is->eof()) {
		t.type = ttEof;
	}
	else {
		eatSpace();
		t.line = line;
		auto ch = is->peek();
		if (isdigit(ch))
			getNumber(t);
		else if (isOperatorWord(ch))
			getOperator(t);
		else
			getLiteral(t);
	}
	return t;
}

void IstreamTokenizer::getLiteral(Token & t)
{
	auto ch = is->get();
	t.str += ch;
	while (!isCutCharactor(ch = is->peek()))
	{
		if (is->eof()) break;
		t.str += ch;
		is->get();
	}
	if (t.str == "end"||t.str == "else"||
		t.str=="if"||t.str=="for"||t.str == "in"
		||t.str =="break"||t.str=="continue"||t.str=="fn"||t.str =="return") {	//ӽ÷ س . ߿ ٲ.
		t.type = ttKeyword;
		return;
	}
	t.type = ttIdentifier;
}

void IstreamTokenizer::getNumber(Token & t)
{
	auto ch = is->get();
	t.str += ch;
	while (isdigit(ch = is->peek())) 
	{
		t.str += ch;
		is->get();
	}
	if (is->peek() == '.') {
		while (isdigit(ch = is->peek()))
		{
			t.str += ch;
			is->get();
		}
		t.type = ttFloat;
	}
	else t.type = ttInteger;
}

void ls::IstreamTokenizer::getOperator(Token & t)
{
	auto ch = is->get();
	t.str += ch;
	while (isOperatorWord(ch = is->peek()))
	{
		t.str += ch;
		is->get();
	}
	t.type = ttOperator;
}

void IstreamTokenizer::eatSpace()
{
	int ch;
	while(isIgnoreWord(ch = is->peek()))
	{
		if (ch == '\n')
			line++;
		is->get();
	}
}

std::istream & ls::operator>>(std::istream & is, Token & t)
{
	IstreamTokenizer tokenizer(is);
	t = tokenizer.get();
	return is;
}
