#include<memory>
#include<unordered_map>
#include<string>
#include"ByteCode.h"
#include"BoxedVar.h"

#ifndef LS_VirtualMachine_H_
#define LS_VirtualMachine_H_
namespace ls {
	//Ƹ  ļ. ߿ ȭ   ִµ غ.
	struct TempVariable{
	private:
		BoxedVar * value;
		BoxedVar temp;
	public:
		explicit TempVariable() :value(nullptr) {}
		explicit TempVariable(BoxedVar * ptr) :value(ptr) {}
		TempVariable(const TempVariable & rhs)
			:value(rhs.value), temp(rhs.temp) {
		}
		inline BoxedVar * val() { 
			if (nullptr == value) return &temp;
			return value;
		}
		inline CompoundType * get() {
			return val()->get().get();
		}
		inline void setval(BoxedVar * ptr) { value = ptr;}
		inline void copyTemp() { 
			temp = *value;
			value = nullptr;
		}

		template<typename Ty>
		Ty * cast(TypeKind t) {
			auto ret = get();
			if (ret->Type() != t)
				throw std::exception("Invaild Type Casting Error!");
			return static_cast<Ty *>(ret);
		}
	};
	class VirtualMachine {
	public:
		VirtualMachine(){
			std::unordered_map<std::string, BoxedVar> t;
			variableMap.push_back(t);
		}
		// . ޸ .
		void set(FunctionCode * ic) { instcodes = ic;
		programFrame = ic;
		programFrameEnd = ic->code.size();
		programCounter = 0;
		}
		bool runOneCode();

		BoxedVar & searchVar(const std::string & str);
		BoxedVar & setVar(const std::string & str) { return variableMap.back()[str]; }
	private:
		void enterScope();
		void leaveScope();
		
		void add();
		void sub();
		void mul();
		void div();
		
		void plus();
		void minus();
		
		void eq();
		void neq();
		void gt();
		void gte();
		void less();
		void leq();
		
		void store();
		void new_var();
		void jmp();
		void jt();
		void jf();
		void call();
		void ref();
		
		void load();

		void and();
		void or();
		void not();

		void pop();
		void push();

		void inst_ret();
		void next();
		void getvalue();
		void isEnd();
		void getiter();

		TempVariable popValue()
		{
			if (stack.size() == 0)
				throw std::runtime_error("invaild");
			auto v = stack.back();
			stack.pop_back();
			return v;
		}
		inline void pushValue(const TempVariable & t) {
			stack.emplace_back(t);
		}
		struct CallStackValue
		{
			FunctionCode * frame;
			pos_t address;
		};
	private:
		std::list<std::unordered_map<std::string, BoxedVar> > variableMap;
		std::vector<TempVariable> stack;
		std::vector<CallStackValue> callstack;
		FunctionCode * instcodes;
		FunctionCode * programFrame;
		pos_t programFrameEnd;
		pos_t programCounter;
	};
}
#endif